import { symbols, Listeners } from './internal';
import { TagNameMap, Attrs, Factory as BaseFactory } from './util/dom';
declare global {
    interface ElementEventMap {
        'mutate': Event;
        'connect': Event;
        'disconnect': Event;
    }
}
export interface El<T extends string = string, E extends Element = Element, C extends El.Children = El.Children> {
    readonly tag: T;
    readonly element: E;
    get children(): El.Getter<C>;
    set children(children: El.Setter<C>);
}
export declare namespace El {
    type Children = Children.Void | Children.Text | Children.Node | Children.Array | Children.Struct;
    namespace Children {
        type Void = void;
        type Text = string;
        type Node = DocumentFragment;
        type Array = readonly El[];
        type Struct = {
            [field: string]: El;
        };
    }
    type Getter<C extends El.Children> = C extends Children.Struct ? C : C;
    type Setter<C extends El.Children> = C extends Children.Struct ? Partial<C> : C;
    type Factory<M extends TagNameMap, C extends El.Children = El.Children> = <T extends keyof M & string>(baseFactory: BaseFactory<M>, tag: T, attrs: Attrs<Extract<M[T], Element>>, children: C) => M[T];
}
export declare class ElementProxy<T extends string = string, E extends Element = Element, C extends El.Children = El.Children> implements El<T, E, C> {
    readonly tag: T;
    readonly element: E;
    constructor(tag: T, element: E, children: C, container?: Element | ShadowRoot);
    private $id;
    private get id();
    private $query;
    private get query();
    private format;
    private isObserverUpdate;
    private observe;
    private readonly type;
    private readonly container;
    private isInit;
    private $children;
    readonly [symbols.listeners]: Listeners;
    get children(): El.Getter<C>;
    set children(children: El.Setter<C>);
}
